############################################################################
#This package include functions that generates basis for sievefuncitonal space
#require(orthopolynom);
############################################################################
#Polynomial
#This function takes input vector x, and degree J_n and returns a
#matrix of polynomial basis of x
sieve.Pol <- function(X, J_n){
    return(unname(outer(X, 0:J_n, "^")));
}
#This function takes input vector x, degree J_n and d (order of derivative)
#then returns a matrix of polynomial basis of x
sieve.Pol.D <- function(X, J_n, d){
    if (d == 1){
        poly_basis_d <- outer(X,0:(J_n-1),"^");
        temp <- t(apply(poly_basis_d,1,function(x)x*(1:J_n)));
        basis <- cbind(0,temp);
        return(unname(basis));
    }
    else if (d==2){
        if (J_n > 1){
        coef_a <- 2:J_n;
        coef_b <- 1:(J_n-1);
        coef <- diag(coef_a%*%t(coef_b));
        poly_basis_dd <- outer(X,0:(J_n-2),"^");
        temp <- t(apply(poly_basis_dd,1,function(x)x*(coef)));
        basis <- cbind(0,0,temp);
        return(basis);
        }
        else{
        n <- length(X);
        return(cbind(rep(0,n),rep(0,n)));
        }
    } #end d==2
    else{
        stop("This function takes only the first and second derivative");
    }
} #end sieve.Pol.D

############################################################################
#Trigonometric polynomials
#This function takes input vector x, and degree J_n and returns trigonometryic
#polynomials on [0,1] of degree J_n
sieve.TriPol <- function(X, J_n){
    if (J_n < 1){
        stop("J_n need to be at least 1");
        }
    basis <- rep(1,length(X));
    for(k in 1:J_n){
       basis <- cbind(basis, cos(2*k*pi*X), sin(2*k*pi*X)); 
    }
    return(unname(basis));
} #sieve.TriPol
#This function takes input vector x, degree J_n and order of derivative d and
#returns the derivative of the trigonometric polynomial basis
sieve.TriPol.D <- function(X, J_n, d){
    if (J_n < 1){
        stop("J_n need to be at least 1");
        }
    basis <- rep(0,length(X));
    if(d==1){
    for(k in 1:J_n){
       basis <- cbind(basis, -2*k*pi*sin(2*k*pi*X), 2*k*pi*cos(2*k*pi*X)); 
    } #end for loop
    } #end d==1
    else{
    for(k in 1:J_n){
       basis <- cbind(basis, -4*k^2*pi^2*cos(2*k*pi*X), 
                             -4*k^2*pi^2*sin(2*k*pi*X)); 
    } #end for loop
    } #end d==1
    return(unname(basis));
} #end sieve.TriPol.D
############################################################################
#Cosine Polynomial
#This function takes input vector x, and degree J_n and returns cosine 
#polyomials on [0,1] of degree J_n
sieve.CosPol <- function(X, J_n){
    if (J_n < 1){
        stop("J_n need to be at least 1");
        }
    basis <- rep(1,length(X));
    for(k in 1:(J_n-1)){
       basis <- cbind(basis, cos(k*pi*X)); 
    }
    return(unname(basis));
} #sieve.CosPol
#This function takes input vector x, degree J_n and order of derivative d and
#returns the derivative of the cosine polynomial basis
sieve.CosPol.D <- function(X, J_n, d){
    if (J_n < 1){
        stop("J_n need to be at least 1");
        }
    basis <- rep(0,length(X));
    if(d==1){
    for(k in 1:(J_n-1)){
       basis <- cbind(basis, -k*pi*sin(k*pi*X));
    } #end for loop
    } #end d==1
    else{
    for(k in 1:(J_n-1)){
       basis <- cbind(basis, -k^2*pi^2*cos(k*pi*X));
    } #end for loop
    } #end d==1
    return(unname(basis));
} #end sieve.CosPol.D
############################################################################
#Sine Polynomial
#This function takes input vector x, and degree J_n and returns cosine 
#polynomials on [0,1] of degree J_n
sieve.SinPol <- function(X, J_n){
    if (J_n < 1){
        stop("J_n need to be at least 1");}
    basis <- sin(pi*X);
    for(k in 2:J_n){
       basis <- cbind(basis, sin(k*pi*X)); 
    }
    return(unname(basis));
} #sieve.SinPol
#This function takes input vector x, degree J_n and order of derivative d and
#returns the derivative of the cosine polynomial basis
sieve.SinPol.D <- function(X, J_n, d){
    if (J_n < 1){
        stop("J_n need to be at least 1");
        }
    if(d==1){
    basis <- pi*cos(pi*X);
    for(k in 2:J_n){
       basis <- cbind(basis, k*pi*cos(k*pi*X));
    } #end for loop
    } #end d==1
    else{
    basis <- -pi^2*sin(pi*X);
    for(k in 2:J_n){
       basis <- cbind(basis, -k^2*pi^2*sin(k*pi*X));
    } #end for loop
    } #end d==1
    return(unname(basis));
} #end sieve.SinPol.D
############################################################################
#Spline
############################################################################
#This functions takes input vector x, a sequence of knots, and the order of
#polynomial of order r and returns a polynomial spline basis matrix for 
#vector X
sieve.SPL <- function(X, r, knots){
    if(r<3){
    stop("We need at least 3rd order, (2nd degree polynomial");
    }
    basis <- outer(X, 0:(r-1), "^");
        for(i in 1:length(knots)){
        basis <- cbind(basis, apply(cbind(X-knots[i],0),1,max)^(r-1));
        } #end i loop for knots
    return(unname(basis));
} #end of sieve.SPL

############################################################################
#This functions takes input vector x, a sequence of knots, the order of
#polynomial p and d order of derivative and returns the 1st or 2nd derivative 
#of the polynomial spline basis matrix for vector X
sieve.SPL.D <- function(X, r, knots, d){
    if(r<3){
    stop("We need at least 3th order, (2nd degree polynomial)");
    }
    if(d>2){
    stop("I can only take the first or second derivative");
    }
    if(d==1){ #first derivative
        basis <- outer(X, 0:(r-2), "^");
        func <- function(x){x*c(1:(r-1))};
        basis <- t(apply(basis, 1, func));
        basis <- cbind(rep(0,length(X)), basis);
        for(i in 1:length(knots)){
            x_knot <- (r-1)*(X-knots[i])^(r-2);
            x_knot[which(X<knots[i])] <- 0;
            basis <- cbind(basis, x_knot);
        } #end loop through knots
        return(unname(basis));
    } #end of d==1
    else{
        if(r==3){
        n = length(X);
        basis <- cbind(rep(0,n),rep(0,n),rep(2,n));
            for(i in 1:length(knots)){
                x_knot <- rep(2,n);
                x_knot[which(X<knots[i])] <-0;
                basis <- cbind(basis, x_knot);
            } #end for loop
        return(unname(basis));
        }
        if(r==4){
        n = length(X);
        basis <- cbind(rep(0,n),rep(0,n),rep(2,n),6*X);
            for(i in 1:length(knots)){
                x_knot <- 6*(X-knots[i]);
                x_knot[which(X<knots[i])] <- 0;
                basis <- cbind(basis, x_knot);
            } #end for loop
        return(unname(basis));
        }
        else{
        n = length(X)
        basis <- outer(X, 1:(r-3), "^");
        d2_coef <- function(x) x*(factorial(3:(r-1))/factorial(1:(r-3)));
        basis <- t(apply(basis, 1, d2_coef));
        basis <- cbind(rep(0,n),rep(0,n),rep(2,n),basis);
            for(i in 1:length(knots)){
                x_knot <- (r-1)*(r-2)*(X-knots[i])^(r-3);
                x_knot[which(X<knots[i])] <- 0;
                basis <- cbind(basis, x_knot);
            } #end for loop
        return(unname(basis));
        } #end else
    } #end second derivative
} #end sieve.SPL.D

############################################################################
#Hermite polynomials
#This function takes a vector of data and output its hermite polynomials
#basis of degree J_n
sieve.HPol <- function(x, J_n){
    poly <- hermite.h.polynomials(J_n);
    basis <- exp(-x^2/2);
    for(j in 2:(J_n+1)){
    ff <- as.function(poly[[j]]);
    basis <- cbind(basis,ff(x)*exp(-x^2/2));
    }
    return(unname(basis));
} #end sieve_HPol
############################################################################
#This function takes a vector of data, and output the first or second
#derivative of the hermite polynomials basis of degree J_n
sieve.HPol.D <- function(x, J_n, d){
    if(d>2){
    stop("only the first and second derivative is allowed");
    } #end d>2
    if(d==1){
    HPol <- hermite.he.polynomials(J_n);
    DHPol<- polynomial.derivatives(HPol);
        for(j in 1:(J_n+1)){
            ff <- as.function(HPol[[j]]);
            Dff <- as.function(DHPol[[j]]);
        if(j==1){
            basis <- (Dff(x) - ff(x)*x)*exp(-x^2/2);
        } #end j==1
        else{
            basis <- cbind(basis, (Dff(x) - ff(x)*x)*exp(-x^2/2));
        } #end else
        } #end j loop
    return(unname(basis));
    } #end if d == 1
    else{
    LPol <- hermite.he.polynomials(J_n);
    DHPol<- polynomial.derivatives(HPol);
    DDHPol <- polynomial.derivatives(DHPol);
        for(j in 1:(J_n+1)){
            ff <- as.function(HPol[[j]]);
            Dff <- as.function(DHPol[[j]]);
            DDff <- as.function(DDHPol[[j]]);
        if(j==1){
            basis <- (DDff(x) - 2*Dff(x)*x - ff(x) + ff(x)*x^2)*exp(-x^2/2);
        } #end j==1
        else{
            basis <- cbind(basis,(DDff(x)-2*Dff(x)*x- ff(x)+ff(x)*x^2)*
                           exp(-x^2/2));
        } #end else
        } #end j loop
    return(unname(basis));
    } #end if d == 2
} #end sieve.HPol.D
############################################################################
#Laguerre polynomials
#This function takes a vector of data and output its laguerre polynomials
#basis of degree J_n
sieve.LPol <- function(x, J_n){
    LPol <- laguerre.polynomials(J_n);
    basis <- exp(-x/2);
    for(j in 2:(J_n+1)){
    ff <- as.function(LPol[[j]]);
    basis <- cbind(basis,ff(x)*exp(-x/2));
    return(unname(basis));
    }
} #end sieve.LPol
############################################################################
#This function takes a vector of data, and output the first or second
#derivative of the hermite polynomials basis of degree J_n
sieve.LPol.D <- function(x, J_n, d){
    if(d>2){
    stop("only the first and second derivative is allowed");
    } #end d>2
    if(d==1){
    LPol <- laguerre.polynomials(J_n);
    DLPol<- polynomial.derivatives(LPol);
        for(j in 1:(J_n+1)){
            ff <- as.function(LPol[[j]]);
            Dff <- as.function(DLPol[[j]]);
        if(j==1){
            basis <- (Dff(x) - ff(x)/2)*exp(-x/2);
        } #end j==1
        else{
            basis <- cbind(basis, (Dff(x) - ff(x)/2)*exp(-x/2));
        } #end else
        } #end j loop
    return(unname(basis));
    } #end if d == 1
    else{
    LPol <- laguerre.polynomials(J_n);
    DLPol<- polynomial.derivatives(LPol);
    DDLPol <- polynomial.derivatives(DLPol);
        for(j in 1:(J_n+1)){
            ff <- as.function(LPol[[j]]);
            Dff <- as.function(DLPol[[j]]);
            DDff <- as.function(DDLPol[[j]]);
        if(j==1){
            basis <- (DDff(x) - Dff(x) + ff(x)/4)*exp(-x/2);
        } #end j==1
        else{
            basis <- cbind(basis, (DDff(x) - Dff(x) + ff(x)/4)*exp(-x/2));
        } #end else
        } #end j loop
    return(unname(basis));
    } #end if d == 2
} #end sieve.LPol.D
############################################################################
#This function takes a matrix B(y) and outputs integral B(y)B(y)' dy
#B(y) is typically the second derivative of the basis of y
D2_basis_int <- function(B){
    k <- dim(B)[2];
    func <- function(x) x%*%t(x);
    C <- apply(B,1,func);
    integral = matrix(apply(C,1,mean),k,k);
    return(unname(integral));
}
############################################################################
